UNIT tren;
(* Programa: TINS'06
 * Unidad: tren
 * Descripcin: Implementa el tren.
 * Autor: uo Martnez <niunio at pagina.de>
 *)

INTERFACE

USES
  mingro;



TYPE
(* ESTADO_LOCOMOTORA:
 *   Mquina de estados de la locomotora. *)
  ESTADO_LOCOMOTORA = (
    LOCOMOTORA_PARADA,	{ Por ejemplo, est en una estacin. }
    LOCOMOTORA_AVANZA,	{ Se desplaza por el mapa... }
    LOCOMOTORA_GIRA	{ Por encontrar una curva o un tope. }
  );



(* LOCOMOTORA:
 *   Maneja el tren, su posicin, su estado... *)
  LOCOMOTORA = CLASS
  PRIVATE
    fImagen: MG_BITMAPptr;
    fEstado: ESTADO_LOCOMOTORA;
    fTiempo: LONGINT; { Un reloj para saber cunto tiempo ha de estar parada. }
    fPantX, fPantY, fAngulo: INTEGER; { Coordenadas en pantalla y ngulo. }
    fAngFin, fAngInc: INTEGER; { ngulo final e incremento de ngulo. }
    fVelocidad: INTEGER;
  PUBLIC
    CONSTRUCTOR Create (Tiles: MG_BITMAPptr);
    DESTRUCTOR Destroy; OVERRIDE;

  (* OcupaCasilla:
   *   Comprueba si la casilla solicitada est ocupada por el tren. *)
    FUNCTION OcupaCasilla (CX, CY: INTEGER): BOOLEAN;
  (* Casilla:
   *   Devuelve las coordenadas de la casilla de cola (-1 si est fuera. *)
    PROCEDURE Casilla (VAR CX, CY: INTEGER);
  (* SiguienteCasilla:
   *   Devuelve las coordenadas de la casilla del morro (-1 si est fuera. *)
    PROCEDURE SiguienteCasilla (VAR CX, CY: INTEGER);
  (* PonTren:
   *   Coloca el tren en su posicin de inicio. *)
    PROCEDURE PonTren (aTX, aTY, aVelocidad: INTEGER);
  (* Actualiza:
   *   Anima la locomotora. *)
    PROCEDURE Actualiza;
  (* Gira:
   *   Le dice a la locomotora que realice un giro. *)
    PROCEDURE Gira (aAngulo: INTEGER; SentidoReloj: BOOLEAN);
  (* Para:
   *   Le dice a la locomotora que se detenga durante unos segundos. *)
    PROCEDURE Para (aSegundos: INTEGER);
  (* Dibuja:
   *   Dibuja el tren en su posicin. *)
    PROCEDURE Dibuja (aBmp: MG_BITMAPptr);

    PROPERTY Estado: ESTADO_LOCOMOTORA READ fEstado;
    PROPERTY PosX: INTEGER READ fPantX;
    PROPERTY PosY: INTEGER READ fPantY;
    PROPERTY Angl: INTEGER READ fAngulo;
  END;



CONST
  FUERA = 255; { Indica que est fuera del tablero. }



IMPLEMENTATION

USES
  error, sysutils;


  
(**************
 * LOCOMOTORA *
 **************)

(* LOCOMOTORA.Create:
 *   Constructor. *)
CONSTRUCTOR LOCOMOTORA.Create (Tiles: MG_BITMAPptr);
BEGIN
  INHERITED Create;
{ Carga el grfico. }
  fImagen := mgCreateSubBitmap (Tiles, 0, 192, 64, 64);
  IF fImagen = NIL THEN
    RAISE EXCEPTION.Create ('Can''t create train!');
  fAngulo := 0;
END;



(* LOCOMOTORA.Destroy:
 *   Destructor. *)
DESTRUCTOR LOCOMOTORA.Destroy;
BEGIN
  IF fImagen <> NIL THEN mgDestroyBitmap (fImagen);
  INHERITED Destroy;
END;



(* LOCOMOTORA.OcupaCasilla:
 *   Comprueba si la casilla solicitada est ocupada por el tren. *)
FUNCTION LOCOMOTORA.OcupaCasilla (CX, CY: INTEGER): BOOLEAN;
VAR
  Ocupa1, Ocupa2: BOOLEAN; { Puede ocupar dos casillas. }
  X, Y: INTEGER;
BEGIN
{ Si est dentro dentro de una casilla, no afecta a las de alrededor. }
  IF ((fPantX MOD 64) = 0) AND ((fPantY MOD 64) = 0) THEN
    OcupaCasilla := ((fPantX DIV 64) = (CX + 1)) AND ((fPantY DIV 64) = (CY + 1))
  ELSE BEGIN { Afecta a dos casillas. }
  { Primera casilla. }
    Casilla (X, Y);
    Ocupa1 := (CX = X) AND (CY = Y);
  { Segunda casilla. }
    IF fAngulo = 0 THEN
      DEC (X);
    IF fAngulo = 64 THEN
      DEC (Y);
    IF fAngulo = 128 THEN
      INC (X);
    IF fAngulo = 192 THEN
      INC (Y);
    Ocupa2 := (CX = X) AND (CY = Y);
    OcupaCasilla := Ocupa1 OR Ocupa2;
  END;
END;



(* LOCOMOTORA.Casilla:
 *   Devuelve las coordenadas de la casilla de cola (-1 si est fuera. *)
PROCEDURE LOCOMOTORA.Casilla (VAR CX, CY: INTEGER);
BEGIN
{ Obtiene la casilla actual. }
  CX := (fPantX DIV 64) - 1; CY := (fPantY DIV 64) - 1;
{ Segn el ngulo, va hacia un sitio u otro. }
  IF (fAngulo <> 128) AND (fAngulo <> 192) THEN
  BEGIN
    IF fAngulo = 0 THEN
      INC (CX);
    IF fAngulo = 64 THEN
      INC (CY);
  END;
END;



(* LOCOMOTORA.SiguienteCasilla:
 *   Devuelve las coordenadas de la casilla del morro (-1 si est fuera. *)
PROCEDURE LOCOMOTORA.SiguienteCasilla (VAR CX, CY: INTEGER);
BEGIN
  CX := -1; CY := -1;
{ Slo ngulos completos, por favor. }
  IF (fAngulo <> 0) AND (fAngulo <> 64) AND (fAngulo <> 128) AND
     (fAngulo <> 192) THEN
    EXIT;
{ Slo si est entre dos casillas. }
  IF (fPantX < 32) OR (fPantY < 32) OR
     (((fPantX MOD 64) <> 0) AND ((fPantX MOD 32) = 0)) OR
     (((fPantY MOD 64) <> 0) AND ((fPantY MOD 32) = 0))
  THEN BEGIN
  { Obtiene la casilla actual. }
    CX := (fPantX DIV 64) - 1; CY := (fPantY DIV 64) - 1;
  { Segn el ngulo, va hacia un sitio u otro. }
    IF (fAngulo <> 0) AND (fAngulo <> 64) THEN
    BEGIN
      IF fAngulo = 128 THEN
	INC (CX);
      IF fAngulo = 192 THEN
	INC (CY);
    END;
  END;
END;



(* LOCOMOTORA.PonTren:
 *   Coloca el tren en su posicin de inicio. *)
PROCEDURE LOCOMOTORA.PonTren (aTX, aTY, aVelocidad: INTEGER);
BEGIN
  fPantX := 64 + (aTX * 64); fPantY := 64 + (aTY * 64);
  fAngulo := 192;
  fEstado := LOCOMOTORA_PARADA; fTiempo := 50;
  fVelocidad := aVelocidad;
END;



(* LOCOMOTORA.Actualiza:
 *   Anima la locomotora. *)
PROCEDURE LOCOMOTORA.Actualiza;
BEGIN
  CASE fEstado OF
  LOCOMOTORA_PARADA: BEGIN
    DEC (fTiempo);
    IF fTiempo = 0 THEN
    BEGIN
      fEstado := LOCOMOTORA_AVANZA;
      SELF.Actualiza; { Para evitar bloqueos al comprobar posicin en ARENA. }
    END;
  END;
  LOCOMOTORA_AVANZA: BEGIN
    IF fAngulo = 0 THEN
      DEC (fPantX, fVelocidad)
    ELSE IF fAngulo = 64 THEN
      DEC (fPantY, fVelocidad)
    ELSE IF fAngulo = 128 THEN
      INC (fPantX, fVelocidad)
    ELSE IF fAngulo = 192 THEN
      INC (fPantY, fVelocidad)
    ELSE
      RAISE EXCEPTION.Create ('Error in train angle!  Can''t be ' +
			      IntTOStr (fAngulo)+'!');
  END;
  LOCOMOTORA_GIRA: BEGIN
    IF fAngulo <> fAngFin THEN
    BEGIN
      INC (fAngulo, fAngInc * fVelocidad);
      IF fAngulo > 255 THEN fAngulo := fAngulo - 256;
      IF fAngulo <   0 THEN fAngulo := fAngulo + 256;
    END
    ELSE
      fEstado := LOCOMOTORA_AVANZA;
  END;
  END;
END;



(* LOCOMOTORA.Gira:
 *   Le dice a la locomotora que realice un giro. *)
PROCEDURE LOCOMOTORA.Gira (aAngulo: INTEGER; SentidoReloj: BOOLEAN);
BEGIN
  fEstado := LOCOMOTORA_GIRA;
  fAngFin := aAngulo;
  IF SentidoReloj THEN
    fAngInc := 2
  ELSE
    fAngInc := -2;
END;



(* LOCOMOTORA.Para:
 *   Le dice a la locomotora que se detenga durante unos segundos. *)
PROCEDURE LOCOMOTORA.Para (aSegundos: INTEGER);
BEGIN
  fEstado := LOCOMOTORA_PARADA;
  fTiempo := aSegundos * 25;
  IF fTiempo = 0 THEN fTiempo := 1;
END;



(* LOCOMOTORA.Dibuja:
 *   Dibuja el tren. *)
PROCEDURE LOCOMOTORA.Dibuja (aBmp: MG_BITMAPptr);
BEGIN
  mgDrawRotatedSprite (aBmp, fImagen, fPantX, fPantY, fAngulo);
END;



END.

